#import "CLLocation.h"

@interface CLLocationInternal : NSObject <NSCoding>
{
	CLLocationCoordinate2D coordinate;
	CLLocationDistance altitude;
	CLLocationAccuracy horizontalAccuracy;
	CLLocationAccuracy verticalAccuracy;
	NSDate* timestamp;
}

@property(readonly, nonatomic) CLLocationCoordinate2D coordinate;
@property(readonly, nonatomic) CLLocationDistance altitude;
@property(readonly, nonatomic) CLLocationAccuracy horizontalAccuracy;
@property(readonly, nonatomic) CLLocationAccuracy verticalAccuracy;
@property(readonly, nonatomic) NSDate* timestamp;

@end

@implementation CLLocationInternal

@synthesize coordinate;
@synthesize altitude;
@synthesize horizontalAccuracy;
@synthesize verticalAccuracy;
@synthesize timestamp;

- (id) initWithCoordinate: (CLLocationCoordinate2D) initCoordinate
				altitude: (CLLocationDistance) initAltitude
				horizontalAccuracy: (CLLocationAccuracy) initHAccuracy
				verticalAccuracy: (CLLocationAccuracy) initVAccuracy
				timestamp: (NSDate*) initTimestamp
{
	self = [super init];
	if (self != nil) 
	{
		coordinate = initCoordinate;
		altitude = initAltitude;
		horizontalAccuracy = initHAccuracy;
		verticalAccuracy = initVAccuracy;
		timestamp = [initTimestamp retain];
	}
	return self;
}

- (void) dealloc
{
	[timestamp release];
	[super dealloc];
}

- (void) encodeWithCoder: (NSCoder*) coder
{
	[coder encodeDouble: coordinate.latitude forKey: @"latitude"];
	[coder encodeDouble: coordinate.longitude forKey: @"longitude"];
	[coder encodeDouble: altitude forKey: @"altitude"];
	[coder encodeDouble: horizontalAccuracy forKey: @"horizontalAccuracy"];
	[coder encodeDouble: verticalAccuracy forKey: @"verticalAccuracy"];
	[coder encodeObject: timestamp forKey: @"timestamp"];
}

- (id) initWithCoder: (NSCoder*) coder
{
	coordinate.latitude = [coder decodeDoubleForKey: @"latitude"];
	coordinate.longitude = [coder decodeDoubleForKey: @"longitude"];
	altitude = [coder decodeDoubleForKey: @"altitude"];
	horizontalAccuracy = [coder decodeDoubleForKey: @"horizontalAccuracy"];
	verticalAccuracy = [coder decodeDoubleForKey: @"verticalAccuracy"];
	timestamp = [coder decodeObjectForKey: @"timestamp"];
	return self;
}
@end

@implementation CLLocation

- (id) initWithCoordinate: (CLLocationCoordinate2D) coordinate
				altitude: (CLLocationDistance) altitude
				horizontalAccuracy: (CLLocationAccuracy) hAccuracy
				verticalAccuracy: (CLLocationAccuracy) vAccuracy
				timestamp: (NSDate*) timestamp
{
	self = [super init];
	if (self != nil) 
	{
		_internal = [[CLLocationInternal alloc] initWithCoordinate: coordinate
																						altitude: altitude
																						horizontalAccuracy: hAccuracy
																						verticalAccuracy: vAccuracy
																						timestamp: timestamp];
	}
	return self;
}

- (id) initWithLatitude: (CLLocationDegrees) latitude longitude: (CLLocationDegrees) longitude
{
	CLLocationCoordinate2D coordinate = {latitude, longitude};
	return [self initWithCoordinate: coordinate altitude: 100.0 horizontalAccuracy: 100.0 verticalAccuracy: 100.0 timestamp:[NSDate date]];
}

- (void) dealloc
{
	[_internal release];
	[super dealloc];
}

- (CLLocationInternal*) internal
{
	return (CLLocationInternal*) _internal;
}

@dynamic coordinate;

- (CLLocationCoordinate2D) coordinate
{ 
	return [self internal].coordinate;
} 

@dynamic altitude;

- (CLLocationDistance) altitude
{ 
	return [self internal].altitude;
}

@dynamic horizontalAccuracy;

- (CLLocationAccuracy) horizontalAccuracy
{
	return [self internal].horizontalAccuracy;
}

@dynamic timestamp;

- (NSDate*) timestamp
{
	return [self internal].timestamp;
}

@dynamic verticalAccuracy;

- (CLLocationAccuracy) verticalAccuracy
{
	return [self internal].verticalAccuracy;
}

- (NSString*) description
{
	return [NSString stringWithFormat:@"<%10.8f, %10.8f> +/- %4.2fm @ %@", [self coordinate].latitude, [self coordinate].longitude, [self horizontalAccuracy], [[self timestamp] description]]; 
}

- (CLLocationDistance) getDistanceFrom: (const CLLocation*) location
{
	const double rad = 57.295779513; // 360 / (2 * pi)
	const double kilconv = 1.853 * 60 * rad;
	
	double distangle = 0.0;
	double lat1 = location.coordinate.latitude;
	double lon1 = location.coordinate.longitude;
	double lat2 = [self coordinate].latitude;
	double lon2 = [self coordinate].longitude;
	const double epsilon = 1e-7;
	lon1 = lon2 - lon1;
	lat1 = (sin(lat1) * sin(lat2) + cos(lat1) * cos(lat2) * cos(lon1));
	if (abs(lat1) > epsilon) 
	{
		lat2 = 1 - lat1 * lat1;
		if ((lat2 > -epsilon) && (lat2 < epsilon))
			lat2 = 0.0;
		lat2 = atan(sqrt(lat2) / lat1);
		if (lat1 < -epsilon)
			lat2 = 3.1415926535 + lat2;
		distangle = lat2 * kilconv;
	}
	else
	{
		distangle = 10006.2; // 1/4 of Earth's circumference, in km
	}
	return distangle;
}

#pragma mark NSCopying

- (id) copyWithZone: (NSZone*) zone
{
    CLLocation* copy = [[CLLocation alloc] initWithCoordinate: [self coordinate] 
																						altitude: [self altitude]
																						horizontalAccuracy: [self horizontalAccuracy] 
																						verticalAccuracy: [self verticalAccuracy]
																						timestamp: [self timestamp]]; 
    return copy;
}

#pragma mark NSCoding

- (void) encodeWithCoder: (NSCoder*) coder
{
	[coder encodeObject: _internal forKey: @"_internal"];
}

- (id) initWithCoder: (NSCoder*) coder
{
	_internal = [coder decodeObjectForKey: @"_internal"];
	return self;
}

@end
