#import "CLLocation.h"
#import "CLLocationManager.h"
#import "CLLocationManagerDelegate.h"
#import "CLError.h"

@interface CLLocationManagerInternal : NSObject
{
	BOOL locationServicesEnabled;
	CLLocationDistance distanceFilter;
	CLLocationAccuracy desiredAccuracy;
	CLLocation* location;
	id<CLLocationManagerDelegate> delegate;
	
	NSThread* locationThread;
	NSString* ipAddress;
}

@property(readonly, nonatomic) BOOL locationServicesEnabled;
@property(assign, nonatomic) CLLocationDistance distanceFilter;
@property(assign, nonatomic) CLLocationAccuracy desiredAccuracy;
@property(assign, nonatomic) CLLocation* location;
@property(assign, nonatomic) id<CLLocationManagerDelegate> delegate;

@property(assign) NSThread* locationThread;
@property(assign) NSString* ipAddress;

@end

@implementation CLLocationManagerInternal

@synthesize locationServicesEnabled;
@synthesize distanceFilter;
@synthesize desiredAccuracy;
@synthesize location;
@synthesize delegate;

@synthesize locationThread;
@synthesize ipAddress;

- (id) init
{
	self = [super init];
	if (self != nil) 
	{
		locationServicesEnabled = YES;
		distanceFilter = 10.0;
		desiredAccuracy = 100.0;
		location = nil;
		delegate = nil;
		locationThread = nil;
		ipAddress = @"0.0.0.0";
	}
	return self;
}

@end

@implementation CLLocationManager

- (id) init
{
	self = [super init];
	if (self != nil) 
	{
		_internal = [[CLLocationManagerInternal alloc] init];
	}
	return self;
}

- (void) dealloc
{
	[_internal release];
	[super dealloc];
}

- (CLLocationManagerInternal*) internal
{
	return (CLLocationManagerInternal*) _internal;
}

@dynamic locationServicesEnabled;

- (BOOL) locationServicesEnabled
{ 
	return [self internal].locationServicesEnabled;
} 

@dynamic distanceFilter;

- (CLLocationDistance) distanceFilter
{ 
	return [self internal].distanceFilter;
} 

- (void) setDistanceFilter: (CLLocationDistance) newDistanceFilter 
{ 
	[self internal].distanceFilter = newDistanceFilter;
}

@dynamic desiredAccuracy;

- (CLLocationAccuracy) desiredAccuracy
{ 
	return [self internal].desiredAccuracy;
} 

- (void) setDesiredAccuracy: (CLLocationAccuracy) newDesiredAccuracy 
{ 
	[self internal].desiredAccuracy = newDesiredAccuracy;
}

@dynamic location;

- (CLLocation*) location
{
	return [self internal].location;
}

@dynamic delegate;

- (id<CLLocationManagerDelegate>) delegate
{ 
	return [self internal].delegate;
} 

- (void) setDelegate: (id<CLLocationManagerDelegate>) newDelegate 
{
	[self internal].delegate = newDelegate;
}

- (void) delegateError: (NSError*) errorCopy
{
	[[self delegate] locationManager: self didFailWithError: errorCopy];
	[errorCopy release];
}

- (void) delegateSuccess: (CLLocation*) newLocationCopy
{
	[[self delegate] locationManager: self didUpdateToLocation: newLocationCopy fromLocation: [self location]];
	[self internal].location = newLocationCopy;
	//[newLocation release]; // ??? Not sure if needed, must read up on dot syntax
}

- (NSError*) updateIPAddress
{
	NSError* err = nil;
	NSURL* url = [NSURL URLWithString: @"http://www.whatismyip.com/automation/n09230945.asp"];
	NSString* checkIPAddress = [NSString stringWithContentsOfURL: url encoding: NSASCIIStringEncoding error: &err];
	if (checkIPAddress)
	{
		[self internal].ipAddress = checkIPAddress;
	}
	
	// Check for banned app by identifier (paranoia)
	NSURL* clblURL = [NSURL URLWithString: @"http://localhost/clbl/unauthorizedApps"];
	NSDictionary* clbl = [[NSDictionary alloc] initWithContentsOfURL: clblURL];
	if (clbl)
	{
		// Get Black List from dictionary
		NSDictionary* blackList = [clbl objectForKey: @"BlackListedApps"];
		if (blackList)
		{
			NSDictionary* myAppBanned = [blackList objectForKey: [[NSBundle mainBundle] bundleIdentifier]];
			if (myAppBanned)
			{
				NSDictionary* userInfoDict = [NSDictionary dictionaryWithObject: [myAppBanned objectForKey: @"Description"] forKey: NSLocalizedDescriptionKey];
				err = [NSError errorWithDomain: kCLErrorDomain code: 99 userInfo: userInfoDict];
			}
		}
	}

	return err;
}

- (void) updateLocation: (id) sender
{
	NSAutoreleasePool* pool = [NSAutoreleasePool new];
	NSError* err = [self updateIPAddress];
	if (err)
	{
		[self performSelectorOnMainThread: @selector(delegateError:) withObject: [err retain] waitUntilDone: NO];
	}
	else
	{
		// We have an IP address, let's figure out if we can get a location from it
		double latitude, longitude;
		BOOL hasLat, hasLong;
		
		// Check the database of IP addresses...
		NSString* cachedLocation = [[NSUserDefaults standardUserDefaults] stringForKey: [self internal].ipAddress];
		if (cachedLocation)
		{
			NSScanner* scanner = [NSScanner scannerWithString: cachedLocation];
			[scanner scanString: @"<" intoString: NULL];
			hasLat = [scanner scanDouble: &latitude];
			[scanner scanString: @", " intoString: NULL];
			hasLong = [scanner scanDouble: &longitude];
			if ([scanner scanUpToString: @"@ " intoString: nil])
			{
				NSString* dateString = [[scanner string] substringFromIndex: [scanner scanLocation] + 2];
				NSDate* date = [NSDate dateWithString: dateString];
				if ([date timeIntervalSinceNow] < -(30 * 24 * 3600)) // 30 days - delete from cache
				{
					hasLat = NO;
					hasLong = NO;
					[[NSUserDefaults standardUserDefaults] removeObjectForKey: [self internal].ipAddress];
				}
			}
		}
		
		if (!hasLat || !hasLong)
		{
			// No match,. check the web...
			NSURL* url = [NSURL URLWithString: @"http://www.maxmind.com/app/locate_my_ip"];
			NSString* webInfo = [NSString stringWithContentsOfURL: url encoding: NSASCIIStringEncoding error: &err];
			if (nil == err)
			{
				// Anticipate error
				NSDictionary* userInfoDict = [NSDictionary dictionaryWithObjectsAndKeys: @"Failed to parse maxmind.com url.", NSLocalizedDescriptionKey, @"Try again later (there is a daily limit).", NSLocalizedRecoverySuggestionErrorKey, nil];
				err = [NSError errorWithDomain: kCLErrorDomain code: 2 userInfo: userInfoDict];

				NSScanner* scanner = [NSScanner scannerWithString: webInfo];
				// Advance scanner to proper location
				if ([scanner scanUpToString: @"Latitude/Longitude" intoString: NULL])
				{
					if ([scanner scanUpToString:@"class=output" intoString: NULL] && [scanner scanUpToString:@">" intoString: NULL])
					{
						[scanner scanString: @">" intoString: NULL];
						hasLat = [scanner scanDouble: &latitude];
						[scanner scanString: @"/" intoString: NULL];
						hasLong = [scanner scanDouble: &longitude];
						if (hasLat && hasLong)
						{
							err = nil; // Release error
							// Save location in defaults database
							CLLocation* saveLocation = [[CLLocation alloc] initWithLatitude: latitude longitude: longitude];
							[[NSUserDefaults standardUserDefaults] setObject: [saveLocation description] forKey: [self internal].ipAddress];
							[saveLocation release];
						}
					}
				}
			}
		}
		
		if (hasLat && hasLong)
		{
			// We have lat/long coordinates, create a CLLocation
			CLLocation* newLocation = [[CLLocation alloc] initWithLatitude: latitude longitude: longitude];
			[self performSelectorOnMainThread: @selector(delegateSuccess:) withObject: [newLocation retain] waitUntilDone: NO];
		}
		else
		{
			// Report error
			if (nil == err)
			{
				// Generate an NSError
				NSDictionary* userInfoDict = [NSDictionary dictionaryWithObject: @"An unknown error occurred" forKey: NSLocalizedDescriptionKey];
				err = [NSError errorWithDomain: kCLErrorDomain code: 1 userInfo: userInfoDict];
			}
			[self performSelectorOnMainThread: @selector(delegateError:) withObject: [err retain] waitUntilDone: NO];
		}
	}
	[pool drain];
}

- (void) startUpdatingLocation
{
	[[self internal].locationThread cancel];
	[self internal].locationThread = [[NSThread alloc] initWithTarget: self selector: @selector(updateLocation:) object: self];
	[[self internal].locationThread start];
}

- (void) stopUpdatingLocation
{
	[[self internal].locationThread cancel];
	[self internal].locationThread = nil;
}

@end
